@@ -302,11 +302,11 @@ static void req_done(struct request *req)
 {
 	int dead = 0;
 
-	list_add(&req->r_wlist, &req->ci->done_reqs);
 	if (conn_tx_on(&req->ci->conn)) {
 		dprintf("connection seems to be dead\n");
 		dead = 1;
-	}
+	} else
+		list_add(&req->r_wlist, &req->ci->done_reqs);
 	client_decref(req->ci);
 
 	if (dead)
@@ -583,11 +583,27 @@ int write_object(struct sheepdog_node_list_entry *e,
 	int i, n, fd, ret, success = 0;
 	char name[128];
 
+	if (nr > nodes)
+		nr = nodes;
+
 	for (i = 0; i < nr; i++) {
 		unsigned rlen = 0, wlen = datalen;
 
 		n = obj_to_sheep(e, nodes, oid, i);
 
+		if (memcmp(e[n].addr, sys->this_node.addr, sizeof(e[n].addr)) == 0 &&
+		    e[n].port == sys->this_node.port) {
+			ret = write_object_local(oid, data, datalen, offset, nr,
+						 node_version, create);
+
+			if (ret != 0)
+				eprintf("fail %"PRIx64" %"PRIx32"\n", oid, ret);
+			else
+				success++;
+
+			continue;
+		}
+
 		addr_to_str(name, sizeof(name), e[n].addr, 0);
 
 		fd = connect_to(name, e[n].port);
@@ -631,6 +647,35 @@ int read_object(struct sheepdog_node_list_entry *e,
 	char name[128];
 	int i = 0, n, fd, ret;
 
+	if (nr > nodes)
+		nr = nodes;
+
+	/* search a local object first */
+	for (i = 0; i < nr; i++) {
+		n = obj_to_sheep(e, nodes, oid, i);
+
+		if (memcmp(e[n].addr, sys->this_node.addr, sizeof(e[n].addr)) == 0 &&
+		    e[n].port == sys->this_node.port) {
+			ret = read_object_local(oid, data, datalen, offset, nr,
+						node_version);
+
+			if (ret < 0) {
+				eprintf("fail %"PRIx64" %"PRId32"\n", oid, ret);
+				return ret;
+			}
+
+			/* fix replication consistency */
+			if (nr > 1) {
+				if (write_object(e, nodes, node_version, oid,
+						 data, datalen, offset, nr, 0))
+					eprintf("failed to fix obj consistency\n");
+			}
+
+			return ret;
+		}
+
+	}
+
 	for (i = 0; i < nr; i++) {
 		unsigned wlen = 0, rlen = datalen;
 
